#  -*- coding: utf-8 -*-
"""
Generator class for generation adequacy assessment

author: Simon Tindemans
email: s.h.tindemans@tudelft.nl

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
import math
import numpy as np
#from numba import jit
import random

import gen_adequacy.randomvar as rv


class Generator(object):
    """Class that represents a single generator or set of identical generators as a linear Markov chain

    """

    def __init__(self, unit_capacity, unit_availability, unit_mtbf, unit_count=1):
        """Initialises a Generator object, consisting of identical units.

        :param unit_capacity: capacity of each unit
        :param unit_availability: availability of each unit [0,1]
        :param unit_mtbf: mean time between failure, in the units of the simulation (usually hours)
        :param unit_count: number of identical units
        """
        assert (unit_capacity > 0), "Generator capacities must be larger than zero."
        assert (0 <= unit_availability <= 1), "Generator availability must be between zero and one."
        assert (unit_mtbf > 0), "Generator MTBF must be positive."
        assert (unit_count >= 1), "Generating unit count must be positive."

        self.unit_capacity = unit_capacity
        self.unit_count = unit_count
        self.unit_availability = unit_availability
        self.unit_MTBF = unit_mtbf

        # Compute the number of units in the 'up' state. This is used for sequential series generation.
        self.up_count = np.random.binomial(unit_count, unit_availability)
        self.unit_fail_rate = 1/(unit_availability * unit_mtbf)
        self.unit_repair_rate = 1/((1 - unit_availability) * unit_mtbf)

    def available_capacity(self):
        """Return sum of capacities of available units.

        :return: sum of capacities of available units.
        """
        return self.up_count * self.unit_capacity

#    @jit
    def power_trace(self, num_steps=1, dt=1.0, random_start=True):
        """
        Return time series of available capacities of the aggregate units.

        :param num_steps: Number of time steps in the series
        :param dt: Size of each time step
        :param random_start: Option to create an independent sequence (True:default) or start from self.up_count (False)
        :return: numpy array of length num_steps and dtype float
        """

        initial_up_count = \
            np.random.binomial(self.unit_count, self.unit_availability) if random_start else self.up_count

        repair_prob = self.unit_repair_rate * dt
        fail_prob = self.unit_fail_rate * dt
        assert repair_prob < 1.0
        assert fail_prob < 1.0

        repair_factor = 1./math.log(1.0-repair_prob)
        fail_factor = 1./math.log(1.0-fail_prob)

        sum_factor = repair_factor + fail_factor

        adjust_trace = np.zeros(num_steps + 1, dtype=int)
        adjust_trace[0] = initial_up_count

        for i in range(self.unit_count):
            if i < initial_up_count:
                state = 1
                change_factor = fail_factor
            else:
                state = -1
                change_factor = repair_factor
            t = 0
            while True:
                t += math.ceil(change_factor * math.log(1.0 - random.random()))
                if t > num_steps:
                    break
                adjust_trace[t] -= state
                state = -state
                change_factor = sum_factor - change_factor

        trace = np.cumsum(adjust_trace)
        self.up_count = trace[-1]

        return self.unit_capacity * trace[:-1].astype(float)

    def rv(self, resolution=1):
        """Returns a random variable representing the available capacity of the generator set.

        :param resolution: desired resolution of the RandomVariable object
        :return: RandomVariable object corresponding to the generator(s) output

        This uses a simple iterative definition for multi-unit generators; this could easily be improved
        (e.g. using a binomial distribution), but is rarely on the critical path.

        If resolution is not a divisor of unit_capacity, the capacity is divided proportionally over
        neighbouring bins, depending on resolution
        """

        # Note that ceil_idx and floor_idx are either identical (if resolution matches) or one apart
        ceil_idx = math.ceil(self.unit_capacity / resolution)
        floor_idx = math.floor(self.unit_capacity / resolution)
        probability_array = np.zeros(int(ceil_idx + 1))

        probability_array[0] = 1.0 - self.unit_availability

        # allocate probability over two identical or neighbouring bins
        distance_to_end = ceil_idx - self.unit_capacity / resolution
        probability_array[floor_idx] += distance_to_end * self.unit_availability
        probability_array[ceil_idx] += (1-distance_to_end) * self.unit_availability

        # define a single unit rv
        unit_rv = rv.RandomVariable(resolution=resolution, base=0, probability_array=probability_array)
        # iterative convolution to determine aggregate rv
        result_rv = rv.RandomVariable(resolution=resolution)
        for i in range(self.unit_count):
            result_rv += unit_rv
        return result_rv



# -*- coding: utf-8 -*-

# Make the following basic classes and functions available from gen_adequacy
from .system import SingleNodeSystem, autogen_system
from .ieee_rts import ieee_rts
from .generator import Generator



#  -*- coding: utf-8 -*-
"""
Discretised random variable functions

author: Simon Tindemans
email: s.h.tindemans@tudelft.nl

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
import math
import numpy as np


class RandomVariable(object):
    """Represents a random variable, defined by a probability mass function on a regular array."""

    def __init__(self, resolution=1, base=0, probability_array=np.array([1.0])):
        """Creates a Random Variable object with probability masses at regular intervals.

        :param resolution: step size of the PMF
        :param base: location of the first probability mass in the array
        :param probability_array: optional PMF array. Values should be positive and sum to 1
        """
        self.base = base
        self.step = resolution
        self.probability_array = np.array(probability_array)
        # compute cumulative distribution function
        self.cdf_array = np.cumsum(self.probability_array)

        self.value_pool = None
        self.value_pool_left = 0
        self.value_pool_random_state = None

    def load_data(self, data_array):
        """Initialise the random variable from an empirical data set.

        This implementation creates a PMF array with bins that are aligned with zero.

        :param data_array: empirical data set (1D array)
        :return: none
        """

        # determine lower and upper bounds for the bins, to compute array size and base
        # TODO: generalise this procedure to distribute probability mass over nearby points (interpolation)
        left_bound = (math.floor(min(data_array)/self.step + 0.5) - 0.5) * self.step
        right_bound = (math.ceil(max(data_array)/self.step - 0.5) + 0.5) * self.step
        num_bins = int(np.round((right_bound - left_bound)/self.step))
        self.base = left_bound + 0.5 * self.step
        # allocate that experimental data across the bins
        prob_array, bins = np.histogram(
            data_array,
            bins=np.linspace(start=left_bound, stop=right_bound, num=(num_bins+1)),
            density=True
        )
        # enforce normalisation as probability mass function, and compute the CDF
        self.probability_array = prob_array * self.step
        self.cdf_array = np.cumsum(self.probability_array)

    def __radd__(self, other):
        return self.__add__(other)

    def __add__(self, other):
        """Overloads the '+' operator.

        Independence is assumed between the variables being added."""
        if type(other) == RandomVariable:
            if self.step != other.step:
                raise NotImplementedError("Cannot add RandomVariable objects with different step sizes.")

            return RandomVariable(
                self.step,
                self.base + other.base,
                np.convolve(self.probability_array, other.probability_array)
            )
        else:
            return RandomVariable(
                self.step,
                self.base + other,
                self.probability_array
            )

    def __neg__(self):
        """Implements the inversion parameter '-'."""
        return RandomVariable(
            base=-(self.base + self.step * (len(self.probability_array) - 1)),
            resolution=self.step,
            probability_array=self.probability_array[::-1]
        )

    def __rsub__(self, other):
        return self.__sub__(other)

    def __sub__(self, other):
        """Overloads the '-' operator.

        Assumes independence between variables."""
        if type(other) == RandomVariable:
            if self.step != other.step:
                raise NotImplementedError("Cannot subtract RandomVariable objects with different step sizes.")

            return RandomVariable(
                self.step,
                self.base - other.base - self.step * (len(other.probability_array)-1),
                np.convolve(self.probability_array, other.probability_array[::-1])
            )
        else:
            return RandomVariable(
                self.step,
                self.base - other,
                self.probability_array
            )

    def __rmul__(self, multiplier):
        return self.__mul__(multiplier)

    def __mul__(self, multiplier):
        """Overloads the '*' operator.

        Only valid for integer values of 'multiplier'."""
        if type(multiplier) == int:
            if multiplier == 0:
                return 0
            elif multiplier > 0:
                new_array = np.zeros(multiplier * (len(self.probability_array)-1) + 1)
                new_array[::multiplier] = self.probability_array
                new_base = self.base * multiplier
            else:
                new_array = np.zeros(- multiplier * (len(self.probability_array)-1) + 1)
                new_array[::multiplier] = self.probability_array
                new_base = multiplier * (self.base + self.step * (len(self.probability_array) - 1))
            return RandomVariable(self.step, base=new_base, probability_array=new_array)
        else:
            raise NotImplementedError("Can only multiply RandomVariable objects by integer values.")

    def __iadd__(self, other):
        """Overloads the '+=' operator."""
        if type(other) == RandomVariable:
            if self.step == other.step:
                self.base += other.base
                self.probability_array = np.convolve(self.probability_array, other.probability_array)
                self.cdf_array = np.cumsum(self.probability_array)
            else:
                raise NotImplementedError("Cannot add RandomVariable objects with different step sizes.")
        else:
            self.base += other
        return self

    def __isub__(self, other):
        """Overloads the '-=' operator."""
        if type(other) == RandomVariable:
            if self.step == other.step:
                self.base -= other.base + self.step * (len(other.probability_array)-1)
                self.probability_array = np.convolve(self.probability_array, other.probability_array[::-1])
                self.cdf_array = np.cumsum(self.probability_array)
            else:
                raise NotImplementedError("Cannot subtract RandomVariable objects with different step sizes.")
        else:
            self.base -= other
        return self

    def __imul__(self, multiplier):
        """Overloads the '*=' operator.

        Is only valid for positive integer values of the multiplier."""
        if type(multiplier) == int:
            if multiplier == 0:
                self.base = 0.0
                self.probability_array = np.numarray([1.0])
            elif multiplier > 0:
                new_array = np.zeros(multiplier * (len(self.probability_array)-1) + 1)
                new_array[::multiplier] = self.probability_array
                self.probability_array = new_array
                self.cdf_array = np.cumsum(self.probability_array)
                self.base = self.base * multiplier
            else:
                new_array = np.zeros(- multiplier * (len(self.probability_array)-1) + 1)
                new_array[::multiplier] = self.probability_array
                self.probability_array - new_array
                self.cdf_array = np.cumsum(self.probability_array)
                self.base = multiplier * (self.base + self.step * (len(self.probability_array) - 1))
        else:
            raise NotImplementedError("Can only multiply RandomVariable objects by integer values.")
        return self

    def random_value(self, number_of_items=1, random_state=None, dither=True):
        """
        Generate random values according to the object's distribution

        :param number_of_items: number of random values to return
        :param random_state: optional numpy RandomState() object to use for the random stream
        :param dither: optional dithering to generate a piecewise-linear PDF (default: True)
        :return: numpy.ndarray() of random values
        """

        res = self.step

        if random_state:
            if dither:
                return random_state.choice(self.x_array(), size=number_of_items, p=self.probability_array) \
                       + random_state.triangular(-res, 0, res, size=number_of_items)
            if not dither:
                return random_state.choice(self.x_array(), size=number_of_items, p=self.probability_array)
        else:
            if dither:
                return np.random.choice(self.x_array(), size=number_of_items, p=self.probability_array) \
                    + np.random.triangular(-res, 0, res, size=number_of_items)
            if not dither:
                return np.random.choice(self.x_array(), size=number_of_items, p=self.probability_array)


    def random_value_from_pool(self, pool_size, random_state=None, dither=True, reset=False):
        """
        Efficient sampler implementation that returns a single value from a pool each time it is called

        :param pool_size: pool size to use when new samples are required
        :param random_state: optional numpy RandomState object to use for number generation
        :param dither: optional dithering to generate a piecewise-linear PDF (default: True)
        :param reset: force repopulation of the pool (default: False)
        :return: random value

        NOTE: the implementation is unsafe, in the sense that a manual reset is required to update the pool values
        each time the system parameters change.
        """
        if reset or (self.value_pool_left <= 0) or random_state != self.value_pool_random_state:
            self.value_pool = self.random_value(number_of_items=pool_size, random_state=random_state, dither=dither)
            self.value_pool_left = pool_size
            self.value_pool_random_state = random_state

        self.value_pool_left = self.value_pool_left - 1
        return self.value_pool[self.value_pool_left]


    def min_max(self):
        """Returns the minimum and maximum values for which probability masses are stored."""
        return self.base, self.base + self.step * (len(self.probability_array)-1)

    def x_array(self):
        """Returns an array of x-values for which probability masses are stored."""
        return np.linspace(*self.min_max(), self.probability_array.size)

    def mean(self):
        """Returns the expectation cdf_value of the probability distribution."""
        return np.dot(self.x_array(), self.probability_array)

    def truncate(self, tolerance=1e-10):
        """Truncates the probability mass function."""
        sum_array = np.cumsum(self.probability_array)
        boundary_indices = np.searchsorted(sum_array, [tolerance, 1.0 - tolerance], side='left')

        new_array = np.copy(self.probability_array[boundary_indices[0]:boundary_indices[1]+1])
        if boundary_indices[0] > 0:
            new_array[0] += sum_array[boundary_indices[0]-1]
        if boundary_indices[1] < len(new_array) - 1:
            new_array[-1] += 1.0 - sum_array[boundary_indices[1]+2]

        self.base += boundary_indices[0] * self.step
        self.probability_array = new_array         # NOTE: copy can be avoided, but this is clearer.
        self.cdf_array = np.cumsum(self.probability_array)

    def quantile(self, q, side='left'):
        """Returns the value associated with the quantile q in [0,1]

        :param q: single cdf_value or list of quantiles to return
        :param side: rounding down ('left') or up ('right').
        :return: returns the type of q
        """

        indices = np.searchsorted(self.cdf_array, q, side=side)
        return self.base + indices * self.step

    def quantile_interpolate(self, q):
        """Returns the value associated with the quantile q in [0,1]

        :param q: single cdf_value or list of quantiles to return
        :return: returns the type of q
        """

        # TODO: create and use interpolated function object.
        # TODO: use numpy.digitize() to perform operation on array indices in one go
        index = np.searchsorted(self.cdf_array, q)  # default 'left' side
        if index != 0:
            delta = (q - self.cdf_array[index - 1]) / (self.cdf_array[index] - self.cdf_array[index - 1])
        else:
            delta = q / self.cdf_array[0]
        return self.base + self.step * (index - 1 + delta + 0.5)

    def cdf_value(self, x):
        """Returns the CDF value [0,1] for a given x

        :param x:
        :return: CDF value [0,1]
        """

        # TODO: vectorise this in x
        if x < self.base:
            return 0.0
        elif x > self.base + self.step * (len(self.cdf_array) - 1):
            return 1.0
        else:
            return self.cdf_array[math.floor((x - self.base)/self.step)]

    def cdf_value_interpolate(self, x):
        """Returns the CDF value [0,1] for a given x

        :param x:
        :return: CDF value [0,1]
        """

        # TODO: vectorise this in x
        x_unit = (x - self.base) / self.step + 0.5
        idx = math.floor(x_unit)
        if idx < 0:
            return 0.0
        if idx >= len(self.cdf_array):
            return 1.0
        else:
            return self.cdf_array[idx] - (idx + 1 - x_unit) * self.probability_array[idx]

    def change_resolution(self, new_resolution):
        """Returns a RandomVariable with new resolution, and distributes the PMF across its bins

        :param new_resolution: new resolution
        """

        # identify lower and upper bounds, and number of bins
        new_base = math.floor(self.base / new_resolution) * new_resolution
        current_pmf_right = self.base + (self.probability_array.size - 1)*self.step
        new_pmf_right = (math.ceil(current_pmf_right / new_resolution)) * new_resolution
        new_num_bins = 1 + int(np.round((new_pmf_right - new_base) / new_resolution))

        # construct probability array
        new_probability_array = np.zeros(new_num_bins)
        for idx, x_value in enumerate(self.x_array()):
            x_unit = (x_value - new_base) / new_resolution
            left_idx = math.floor(x_unit)
            right_idx = math.ceil(x_unit)
            distance_from_end = right_idx - x_unit
            new_probability_array[left_idx] += distance_from_end * self.probability_array[idx]
            new_probability_array[right_idx] += (1 - distance_from_end) * self.probability_array[idx]

        return RandomVariable(
            resolution=new_resolution,
            base=new_base,
            probability_array=new_probability_array
        )



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
"""
Class and supporting function to instantiate a single node generation adequacy problem.

author: Simon Tindemans
email: s.h.tindemans@tudelft.nl

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
import math
import numpy as np
#from numba import jit

import gen_adequacy.randomvar as rv
from gen_adequacy.generator import Generator


# TODO: fix smoothing in epns()

class SingleNodeSystem(object):
    """Class representing generic single bus system with two-state generators, and profiles for wind power and load."""

    def __init__(self, gen_list, load_profile, wind_profile=None, resolution=1, load_offset=None):
        """Initialises the System object with a supplied list of generators and a load profile.

        :param gen_list: list of Generator objects
        :param load_profile: array with load profile
        :param wind_profile: array with wind profile. Assumed independent from load.
        :param resolution: resolution for random variables
        :param load_offset: structural load offset

        It is assumed that a system is instantiated only once and is not modified afterwards.
        """
        self.resolution = resolution
        self.gen_list = gen_list
        self.load_profile = load_profile
        self.load_offset = load_offset

        # NOTE: wind_profile may be None
        self.wind_profile = wind_profile

    @property
    def generation_rv(self):
        """Return a random variable representing the available generation of 'areas' RTS areas.

        :return: RandomVariable object that represents the generating capacity
        """
        if not hasattr(self, '_gen_rv'):
            self._gen_rv = rv.RandomVariable(resolution=self.resolution)
            for generator in self.gen_list:
                self._gen_rv += generator.rv(resolution=self.resolution)
        return self._gen_rv

    @property
    def load_rv(self):
        """Return a random variable representing the load duration curve of 'areas' RTS areas.

        :return: RandomVariable object that represents the system load
        """
        if not hasattr(self, '_ld_rv'):
            self._ld_rv = rv.RandomVariable(resolution=self.resolution)
            self._ld_rv.load_data(self.load_profile)
            if self.load_offset is not None:
                self._ld_rv += self.load_offset
        return self._ld_rv

    @property
    def wind_rv(self):
        """Return a random variable representing the gen duration curve of wind power.

        :return: RandomVariable object that represents the system load
        """
        if self.wind_profile is None:
            return None
        if not hasattr(self, '_w_rv'):
            self._w_rv = rv.RandomVariable(resolution=self.resolution)
            self._w_rv.load_data(self.wind_profile)
        return self._w_rv

    @property
    def margin_rv(self):
        """Return a random variable representing the generation margin.

        :return: RandomVariable object that represents the system margin
        """
        if not hasattr(self, '_m_rv'):
            if self.wind_profile is None:
                self._m_rv = self.generation_rv - self.load_rv
            else:
                self._m_rv = self.generation_rv + self.wind_rv - self.load_rv
        return self._m_rv

    def lolp(self, load_offset=0, interpolation=True):
        """
        Compute system LOLP (loss of load probability).

        :param load_offset: Additional load added to the system prior to computing LOLP
        :param interpolation: Specify whether linear interpolation of probability mass is used.
        :return: LOLP value in range [0,1]
        """
        if interpolation:
            return self.margin_rv.cdf_value_interpolate(load_offset)
        else:
            return self.margin_rv.cdf_value(load_offset - 1E-10)

    def lole(self, load_offset=0):
        """
        Compute system LOLE (loss of load expectation).

        :param load_offset: Additional load added to the system prior to computing LOLE
        :return: LOLE value

        The LOLE is computed by summing loss-of-load probabilities for each load level in self.load_profile. If an
        hourly annual load profile is used, this results in LOLE (hr/year); when a profile of daily peak loads is used,
        it returns LOLE (days/year).

        Note that LOLE is approx. LOLP * len(load_profile), but there is a small numerical difference due to the fact
        that LOLP is based on the margin distribution, which uses load levels that have been discretised onto a grid.
        As a result, the lole() result should be used for comparison with LOLE values in the literature.
        """
        total_load_offset = (0 if self.load_offset is None else self.load_offset) + load_offset
        if self.wind_rv is None:
            gw_rv = self.generation_rv - total_load_offset
        else:
            gw_rv = self.generation_rv + self.wind_rv - total_load_offset
        return np.sum([gw_rv.cdf_value(load_level - 1E-10) for load_level in self.load_profile])

    def epns(self, load_offset=0, interpolation=True):
        """
        Compute EPNS (expected power not supplied).

        :param load_offset: Additional load added to the system prior to computing EPNS
        :param interpolation: Specify whether linear interpolation of probability mass is used.
        :return: EPNS value
        """
        # TODO: implement interpolation
        mrv = self.margin_rv
        # determine index corresponding to target offset (on left)
        ceil_idx = max(0, min(math.ceil((load_offset - mrv.base)/mrv.step), len(mrv.cdf_array)))

        return mrv.step * mrv.cdf_array[0:ceil_idx].sum() + \
            mrv.cdf_array[ceil_idx] * (load_offset - mrv.base - ceil_idx * mrv.step)

    def compute_load_offset(self, lolp_target):
        """
        Determine additional load that must be added to the system to achieve the target LOLP.

        :param lolp_target: LOLP target in [0,1]
        :return: additional load level

        The load addition is implemented as a constant addition (no profile). It is effectively the inverse of
        lolp_target = lolp(load, interpolation=True). Interpolation is used to ensure a unique solution.
        """
        return self.margin_rv.quantile_interpolate(lolp_target)

    def margin_sample(self, number_of_values=1):
        """
        Generate random samples from the margin distribution.

        :param number_of_values: number of samples
        :return: single value or numpy array of values
        """
        mrv = self.margin_rv
        return mrv.random_value(number_of_items=number_of_values)

#    @jit
    def generation_trace(self, num_steps=0, dt=1.0, random_start=True):
        """
        Sample a random generation trace for all generators in the system.

        :param num_steps: Number of steps
        :param dt: Time step
        :param random_start: Whether to create an independent sequence (True; default) or start from previous state (False)
        :return: Sequence of available capacity values
        """

        # infer length of trace
        if num_steps == 0:
            generate_steps = len(self.load_profile)
            if self.wind_profile is not None:
                generate_steps = max(generate_steps, len(self.wind_profile))
        else:
            generate_steps = num_steps

        trace = np.zeros(generate_steps)
        for gen in self.gen_list:
            trace += gen.power_trace(num_steps=generate_steps, dt=dt, random_start=random_start)

        return trace


# TODO: pass on most arguments to make less fragile
def autogen_system(
        load_profile=None, peak_load=None, wind_profile=None, resolution=10,
        gen_availability=0.90,  MTBF=2000, LOLH=3, base_unit=None, max_unit=None, gen_set=None, num_sets=None,
        apply_load_offset=False):
        """Initialises the generators and load curve of the RTS object and calls the System initialiser.

        :param load_profile: optional numpy.ndarray load profile of arbitrary length
        :param peak_load: if specified, peak load of the system, in same units as generator capacities
        :param wind_profile: optinoal numpy.ndarray wind profile of arbitrary length. Assumed independent from load.
        :param resolution: resolution for the random variables
        :param gen_availability: availability of each generator
        :param MTBF: generator MTBF (not required for steady state analysis)
        :param LOLH: system risk standard (expected loss-of-load hours)
        :param base_unit: size of smallest unit in the system
        :param max_unit: upper bound for maximum unit size. Actual unit sizes are generated in powers of two from base_unit
        :param gen_set: array of generator sizes
        :param num_sets: number of generator sets to use ('None' for automatic optimisation)
        :param apply_load_offset: adjust load profile to hit LOLH target exactly
        """

        assert (max_unit is None and base_unit is None) or (gen_set is None)

        if gen_set is not None:
            gen_types = gen_set
        else:
            assert (max_unit >= base_unit), "autogen_system: max_unit should exceed base_unit"

            # Create a 'basic' generator set, starting from base_unit (in MW), increasing by factors of two until 'max_unit'
            gen_types = []
            current_size = base_unit
            while current_size <= max_unit:
                gen_types.append(current_size)
                current_size *= 2
        print('Generator set:', gen_types)


        # Check whether a load_profile has been supplied. If not, use a constant peak_load
        if load_profile is None:
            assert peak_load is not None
            load_profile = np.array([1.])
        else:
            assert isinstance(load_profile, (np.ndarray, np.generic))

        # Check whether peak_load has been supplied. If so, rescale load profile to target peak load
        if peak_load is not None:
            load_profile *= peak_load / load_profile.max()

        # Check whether the number of generator sets must be optimised.
        if num_sets is None:

            print("Optimising number of generator sets to LOLE target of", LOLH, "hours")

            # initialise generation portfolio
            gen_list = [
                Generator(unit_count=1, unit_capacity=power, unit_availability=gen_availability, unit_mtbf=MTBF)
                for (idx, power) in enumerate(gen_types)]
            num_sets = 0

            # create temporary load and gen rv's for iteration
            net_ld_rv = rv.RandomVariable(resolution=resolution)
            net_ld_rv.load_data(load_profile)

            if wind_profile is not None:
                wind_rv = rv.RandomVariable(resolution=resolution)
                wind_rv.load_data(wind_profile)
                net_ld_rv -= wind_rv

            gen_rv = rv.RandomVariable(resolution=resolution)

            # add sets of generators one by one, until the risk is less than the threshold
            while True:
                for generator in gen_list:
                    gen_rv += generator.rv(resolution=resolution)

                gen_rv.truncate()
                num_sets += 1

                if (gen_rv - net_ld_rv).cdf_value(-0.00001) * 8760 < LOLH:
                    break

            # Report number of generator sets and LOLH value
            print("{num} generator sets. Base LOLE={lole:.3g} h; determined load offset of {load:.4g} MW to reach LOLE target.\n".
                  format(num=num_sets,
                         lole=(gen_rv - net_ld_rv).cdf_value_interpolate(0.0) * 8760,
                         load=(gen_rv - net_ld_rv).quantile_interpolate(LOLH / 8760)
                         )
                  )

            if apply_load_offset is True:
                load_offset = (gen_rv - net_ld_rv).quantile_interpolate(LOLH/8760)
            else:
                load_offset = None

        else:
            load_offset = None

        # Generate generator list
        gen_list = [
            Generator(unit_count=num_sets, unit_capacity=power, unit_availability=gen_availability, unit_mtbf=MTBF)
            for (idx, power) in enumerate(gen_types)]

        # initialise parent object
        return SingleNodeSystem(gen_list=gen_list, load_profile=load_profile,
                                wind_profile=wind_profile, resolution=resolution,
                                load_offset=load_offset)


if __name__ == "__main__":
    print("Starting direct execution\n")

    system = autogen_system(load_profile=None, peak_load=10000, wind_profile=None, resolution=10,
                            gen_availability=0.90,  MTBF=2000, LOLH=3, base_unit=10, max_unit=1000, gen_set=None,
                            apply_load_offset=True)

    print("Properties after adjustment:")
    print("LOLE: {:.4f} h".format(8760 * system.lolp()))
    print("EPNS: {:.4g} h".format(system.epns()))



from setuptools import setup

# Based on tutorial https://packaging.python.org/tutorials/packaging-projects/

# Get the long description from the README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='gen_adequacy',
    version='0.4.0',
    packages=['gen_adequacy'],
    url='https://github.com/simontindemans/gen_adequacy',
    author='Simon Tindemans',
    author_email='s.h.tindemans@tudelft.nl',
    description='Tools for single node generation adequacy analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
    ],
    install_requires=['numpy', 'numba'],
    python_requires='>=3.5',
)



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
"""
IEEE RTS Generation System

author: Simon Tindemans
email: s.h.tindemans@tudelft.nl

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
import numpy as np

from gen_adequacy.system import SingleNodeSystem
from gen_adequacy.generator import Generator

# NOTE: the __init__ file imports ieee_rts directly into the gen_adequacy namespace


def ieee_rts(resolution=1, areas=1):
        """Initialises the generators and load curve of the RTS object and calls the GenLoadSystem initialiser.

        :param resolution: desired resolution for any random variables (generation or load)
        :param areas: number of RTS areas
        """

        # Initialise generation portfolio
        gen_list = [
            Generator(unit_count=5 * areas, unit_capacity=12, unit_availability=0.98, unit_mtbf=3000),
            Generator(unit_count=4 * areas, unit_capacity=20, unit_availability=0.9, unit_mtbf=500),
            Generator(unit_count=6 * areas, unit_capacity=50, unit_availability=0.99, unit_mtbf=2000),
            Generator(unit_count=4 * areas, unit_capacity=76, unit_availability=0.98, unit_mtbf=2000),
            Generator(unit_count=3 * areas, unit_capacity=100, unit_availability=0.96, unit_mtbf=1250),
            Generator(unit_count=4 * areas, unit_capacity=155, unit_availability=0.96, unit_mtbf=1000),
            Generator(unit_count=3 * areas, unit_capacity=197, unit_availability=0.95, unit_mtbf=1000),
            Generator(unit_count=1 * areas, unit_capacity=350, unit_availability=0.92, unit_mtbf=1250),
            Generator(unit_count=2 * areas, unit_capacity=400, unit_availability=0.88, unit_mtbf=1250)
        ]

        weekly_load_factors = np.array([86.2, 90.0, 87.8, 83.4, 88.0, 84.1, 83.2, 80.6, 74.0, 73.7,
                                        71.5, 72.7, 70.4, 75.0, 72.1, 80.0, 75.4, 83.7, 87.0, 88.0,
                                        85.6, 81.1, 90.0, 88.7, 89.6, 86.1, 75.5, 81.6, 80.1, 88.0,
                                        72.2, 77.6, 80.0, 72.9, 72.6, 70.5, 78.0, 69.5, 72.4, 72.4,
                                        74.3, 74.4, 80.0, 88.1, 88.5, 90.9, 94.0, 89.0, 94.2, 97.0,
                                       100.0, 95.2]) / 100
        daily_load_factors = np.array([93, 100, 98, 96, 94, 77, 75]) / 100
        winter_weekday_hourly = np.array([67, 63, 60, 59, 59, 60,
                                          74, 86, 95, 96, 96, 95,
                                          95, 95, 93, 94, 99, 100,
                                         100, 96, 91, 83, 73, 63]) / 100
        winter_weekend_hourly = np.array([78, 72, 68, 66, 64, 65,
                                          66, 70, 80, 88, 90, 91,
                                          90, 88, 87, 87, 91, 100,
                                          99, 97, 94, 92, 87, 81]) / 100
        summer_weekday_hourly = np.array([64, 60, 58, 56, 56, 58,
                                          64, 76, 87, 95, 99, 100,
                                          99, 100, 100, 97, 96, 96,
                                          93, 92, 92, 93, 87, 72]) / 100
        summer_weekend_hourly = np.array([74, 70, 66, 65, 64, 62,
                                          62, 66, 81, 86, 91, 93,
                                          93, 92, 91, 91, 92, 94,
                                          95, 95, 100, 93, 88, 80]) / 100
        spring_fall_weekday_hourly = np.array([63, 62, 60, 58, 59, 65,
                                               72, 85, 95, 99, 100, 99,
                                               93, 92, 90, 88, 90, 92,
                                               96, 98, 96, 90, 80, 70]) / 100
        spring_fall_weekend_hourly = np.array([75, 73, 69, 66, 65, 65,
                                               68, 74, 83, 89, 92, 94,
                                               91, 90, 90, 86, 85, 88,
                                               92, 100, 97, 95, 90, 85]) / 100

        winter_week = np.concatenate((np.tile(winter_weekday_hourly, (5,)), np.tile(winter_weekend_hourly, (2,)))) \
            * np.repeat(daily_load_factors, 24)
        summer_week = np.concatenate((np.tile(summer_weekday_hourly, (5,)), np.tile(summer_weekend_hourly, (2,)))) \
            * np.repeat(daily_load_factors, 24)
        spring_fall_week = np.concatenate((np.tile(spring_fall_weekday_hourly, (5,)),
                                           np.tile(spring_fall_weekend_hourly, (2,)))) \
            * np.repeat(daily_load_factors, 24)

        load_profile = areas * 2850 * np.repeat(weekly_load_factors, 7*24) * np.concatenate((
            np.tile(winter_week, (8,)),
            np.tile(spring_fall_week, (9,)),
            np.tile(summer_week, (13,)),
            np.tile(spring_fall_week, (13,)),
            np.tile(winter_week, (9,))
        ))

        # initialise parent object
        return SingleNodeSystem(gen_list=gen_list, load_profile=load_profile, resolution=resolution)


if __name__ == "__main__":
    print("Starting direct execution")

    rts1 = ieee_rts(areas=1, resolution=1)
    rts3 = ieee_rts(areas=3)

    print("\nIEEE RTS 1 Area:")
    print("LOLP [discrete levels]       : {:.6g}".format(rts1.lolp(interpolation=False)))
    print("LOLP [linear interpolation]  : {:.6g}".format(rts1.lolp(interpolation=True)))
    print("EENS [discrete levels]       : {:.6g} MWh".format(rts1.epns(interpolation=False) * 8736))
    print("EENS [linear interpolation]  : {:.6g} MWh".format(rts1.epns(interpolation=True) * 8736))
    print("LOLE                         : {:.6g} h".format(rts1.lole()))
    print("Load offset for LOLP==3/8736 : {:.6g} MW".format(rts1.compute_load_offset(lolp_target=3/8736)))

    print("\nIEEE RTS 3 Area (single node representation):")
    print("LOLP                         : {:.6g}".format(rts3.lolp()))
    print("EENS                         : {:.6g} MWh".format(rts3.epns() * 8736))
    print("LOLE                         : {:.6g} h".format(rts3.lole()))
    print("Load offset for LOLP==3/8736 : {:.6g} MW".format(rts3.compute_load_offset(lolp_target=3/8736)))
